from typing import Dict, Any
import json
from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator
from airflow.models import Variable
from airflow.decorators import task

from components.Utils import select_multiple
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from components.DatabaseOperation.DatabaseTransformation import delete_table
from components.S3Route import load_control_to_s3
from enums.OperationTypeEnum import OperationTypeEnum

import logging

logger = logging.getLogger()


def validate_clean(control_params: Dict[str, Any], provider: str, timezone: str, **kwargs) -> None:
    delete_task_instances()
    ti = kwargs["ti"]
    conf = ti.xcom_pull(task_ids="VALIDATE_GENERATOR", key="CONTROL-CONFIG")
    conn = control_params["connection_id"]
    bucket = control_params["bucket"]
    prefix = control_params["prefix"]
    if not prefix.endswith("/"):
        prefix += "/"
    key = prefix + control_params["filename"]
    conf = json.dumps(conf, indent=2, default=str)
    loaded = load_control_to_s3(bytes(conf.encode()), conn, bucket, key, provider, timezone)
    if loaded:
        logger.info(f"Cargado correctamente el archivo de control en {key}")
    delete_all_xcom_tasks()


def clean(command: str, intern_conn):
    engine = intern_conn.engine
    tablename = select_multiple(command)["tablename"]
    logger.info(f"Borrando tabla {tablename}")
    delete = delete_table(tablename, engine)
    if delete:
        logger.info(f"Borrado correctamente la tabla {tablename}")
    logger.info(f"Borrado todas las variables xcom")


@task(task_id="MASTER_CLEANING", trigger_rule='none_skipped')
def get_cleaners_from_xcom(**kwargs):
    final_selects = []
    task = kwargs['ti']
    xcom_keys = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="XCOM-EXTRACTION-NAMES")
    logger.debug(xcom_keys)
    for key in xcom_keys:
        if not key.startswith(OperationTypeEnum.SELECT.value) and not key.startswith(OperationTypeEnum.PROCEDURE.value):
            continue
        xcom_outputs = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key=key)
        logger.info(f"Trayendo comandos {xcom_outputs}")
        for select in xcom_outputs:
            final_selects.append(select)
    logger.info(f"Final selects: {final_selects}")
    Variable.set(key='CLEANS', value=final_selects, serialize_json=True)
    if len(final_selects) > 0:
        return [[item] for item in final_selects]
    else:
        return [[None]]


def get_cleaning_task_group(db_intern_conn, control_s3: Dict[str, Any], provider: str,
                            timezone: str) -> TaskGroup or None:
    group = None
    try:
        with TaskGroup(group_id="LimpiezaDelProceso", prefix_group_id=False) as group:
            cleaners = get_cleaners_from_xcom()

            tasks = PythonOperator.partial(
                task_id="CLEANERS",
                python_callable=clean,
                op_kwargs={'intern_conn': db_intern_conn}
            ).expand(op_args=cleaners)

            validate_task = PythonOperator(
                task_id="VALIDATE_CLEANER",
                python_callable=validate_clean,
                op_kwargs={'control_params': control_s3, 'provider': provider, 'timezone': timezone},
                trigger_rule='none_skipped'
            )
            cleaners >> tasks >> validate_task
    except Exception as e:
        logger.error(f"Error creando taskGroup de limpiadores. {e}")
    finally:
        return group
