from typing import Any, Dict
import os
import json
from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator
from airflow.models import Variable
from airflow.decorators import task
from airflow.exceptions import AirflowSkipException

from enums.ProcessStatusEnum import ProcessStatusEnum
from components.S3Route import save_df_to_s3, load_control_to_s3
from components.Utils import select_multiple, create_temp_file, delete_temp_dir
from components.Control import get_tasks_from_control, update_new_process
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from components.DatabaseOperation.DatabaseExtraction import get_iterator, get_steps
from enums.OperationTypeEnum import OperationTypeEnum

import logging

logger = logging.getLogger()


def validate_generate(control_params: Dict[str, Any], timezone: str, provider: str, **kwargs) -> None:
    delete_task_instances()
    ti = kwargs["ti"]
    success_tasks = ti.xcom_pull(task_ids="GENERATORS", key="SUCCESS_TASKS")
    failed_tasks = ti.xcom_pull(task_ids="GENERATORS", key="FAILED_TASKS")
    conf = ti.xcom_pull(task_ids="VALIDATE_TRANSFORMATION", key="CONTROL-CONFIG")
    final_dict = {}
    status = ProcessStatusEnum.SUCCESS.value
    if not isinstance(success_tasks, type(None)) and len(success_tasks) > 0:
        for success_task in success_tasks:
            task = ti.xcom_pull(task_ids="GENERATORS", key=success_task)[0]
            final_dict.update({success_task: task})
    if not isinstance(failed_tasks, type(None)) and len(failed_tasks) > 0:
        status = ProcessStatusEnum.FAIL.value
        for failed_task in failed_tasks:
            task = ti.xcom_pull(task_ids="GENERATORS", key=failed_task)[0]
            final_dict.update({failed_task: task})
    conf = update_new_process(conf, status, final_dict, timezone, ti, True)
    if status == ProcessStatusEnum.FAIL.value:
        conn = control_params["connection_id"]
        bucket = control_params["bucket"]
        prefix = control_params["prefix"]
        if not prefix.endswith("/"):
            prefix += "/"
        key = prefix + control_params["filename"]
        conf = json.dumps(conf, indent=2, default=str)
        loaded = load_control_to_s3(bytes(conf.encode()), conn, bucket, key, provider, timezone)
        if loaded:
            logger.info(f"Cargado correctamente el archivo de control en {key}")
        delete_all_xcom_tasks()
        raise AirflowSkipException(f"Ocurrieron errores en la etapa de generación")
    elif status == ProcessStatusEnum.SUCCESS.value:
        ti.xcom_push(key="CONTROL-CONFIG", value=conf)


def on_failure_generator(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('GENERATES', default_var=[], deserialize_json=True)
    table = selects[ti.map_index]
    table = select_multiple(table)["tablename"]
    exception = str(context["exception"])
    status = ProcessStatusEnum.FAIL.value
    task_result = {"description": table, "status": status, "message": exception}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="FAILED_TASKS", value=task_name)


def on_success_generator(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('GENERATES', default_var=[], deserialize_json=True)
    table = selects[ti.map_index]
    table = select_multiple(table)["tablename"]
    status = ProcessStatusEnum.SUCCESS.value
    task_result = {"description": table, "status": status, "message": ""}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="SUCCESS_TASKS", value=task_name)


def generate_and_deploy(command: str, intern_conn, params: Dict[str, Any], timezone: str,
                        provider: str, chunksize=10000):
    if isinstance(command, type(None)):
        raise AirflowSkipException
    engine = intern_conn.engine
    logger.debug(f"COMANDO: {command}")
    tablename = select_multiple(command)["tablename"]
    logger.info(f"Generando resultado de la tabla {tablename}")
    # Creamos el archivo temporal
    filename_mask = params["filename_mask"]
    file_type = params["file_type"]
    pattern = params["datetime_pattern"]
    delimiter = params["delimiter"]
    tmp_path = params["tmp_path"]
    tmp_file = create_temp_file(tmp_path, filename_mask, file_type, tablename, timezone, pattern)
    logger.info(tmp_file)
    steps = get_steps(tablename, chunksize, engine, True)
    iterator = get_iterator(tablename, chunksize, engine)
    logger.info(f"Total de pasos para generar archivo resultado: {steps}")
    for step in range(steps):
        logger.debug(f"STEP: {step}")
        header = True if step == 0 else False
        try:
            dataframe = next(iterator)
            dataframe = dataframe.drop("INTERN_ID_BCOM", axis=1, errors='ignore')
            logger.debug(dataframe)
            dataframe.to_csv(tmp_file, sep=delimiter, index=False, mode='a', header=header)
        except StopIteration:
            break

    bucket = params["s3_params"]["bucket"]
    prefix = params["s3_params"]["prefix"]
    conn_id = params["s3_params"]["connection_id"]
    if not prefix.endswith("/"):
        prefix += "/"
    file_key = prefix + tmp_file[tmp_file.rfind("/")+1:]
    # Se sube el archivo al S3
    logger.info(f"Tamaño del archivo a subir: {os.path.getsize(tmp_file)} bytes")
    save_df_to_s3(tmp_file, conn_id, bucket, file_key, provider, in_memory=False)

    # Se borra el archivo al finalizar el upload
    delete_temp_dir(tmp_file)


@task(task_id="MASTER_GENERATION", trigger_rule='none_skipped')
def get_generate_from_xcom(**kwargs):
    task = kwargs['ti']
    final_outputs = []
    conf = task.xcom_pull(task_ids="VALIDATE_TRANSFORMATION", key="CONTROL-CONFIG")
    tasks = get_tasks_from_control(conf, "generator")
    success_tasks = tasks["tasks"]
    success_tasks = [item[1] for item in success_tasks]
    logger.info(f"GENERADORES QUE FUERON EXITOSOS (TABLAS): {success_tasks}")
    xcom_keys = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="XCOM-EXTRACTION-NAMES")
    logger.debug(xcom_keys)
    for key in xcom_keys:
        if not key.startswith(OperationTypeEnum.SELECT.value) and not key.startswith(OperationTypeEnum.PROCEDURE.value):
            continue
        xcom_outputs = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key=key)
        logger.info(f"Trayendo tablas {xcom_outputs}")
        for select in xcom_outputs:
            if tasks["reset"] or tasks["status"] == ProcessStatusEnum.SUCCESS.value or select not in success_tasks:
                final_outputs.append(select)
    logger.info(f"Final outputs: {final_outputs}")
    Variable.set(key='GENERATES', value=final_outputs, serialize_json=True)
    if len(final_outputs) > 0:
        return [[item] for item in final_outputs]
    else:
        return [[None]]


def get_generate_task_group(db_intern_conn, parameters: Dict[str, Any], control_s3: Dict[str, Any],
                            timezone: str, provider: str) -> TaskGroup or None:
    group = None
    try:
        with TaskGroup(group_id="GeneracionyDespliegueDeResultados", prefix_group_id=False) as group:
            outputs = get_generate_from_xcom()

            tasks = PythonOperator.partial(
                task_id="GENERATORS",
                python_callable=generate_and_deploy,
                on_failure_callback=on_failure_generator,
                on_success_callback=on_success_generator,
                op_kwargs={'intern_conn': db_intern_conn, 'params': parameters, 'timezone': timezone,
                           'provider': provider}
            ).expand(op_args=outputs)

            validate_task = PythonOperator(
                task_id="VALIDATE_GENERATOR",
                python_callable=validate_generate,
                op_kwargs={'control_params': control_s3, 'timezone': timezone, 'provider': provider},
                trigger_rule='none_skipped'
            )
            outputs >> tasks >> validate_task
    except Exception as e:
        logger.error(f"Error creando taskGroup de generadores. {e}")
    finally:
        return group
