from datetime import datetime
from typing import Any, Dict

from airflow import DAG
import pandas as pd
import os
import uuid
from airflow.operators.python import PythonOperator

from components.Control import extract_last_control
from components.S3Route import load_report_to_s3
from enums.ProcessStatusEnum import ProcessStatusEnum
from components.Utils import delete_temp_dir

import logging

logger = logging.getLogger()


DAG_NAME = "INFORM_PROCESS"

# Change this path if is deployed in prod or dev
MAIN_PATH = "/root/airflow/dags/"

DEFAULT_ARGS = {
    'owner': 'BCOM',
    "start_date": datetime(2023, 7, 29, 22, 9),
    'depends_on_past': False,
    'email': 'caguirre@bytesw.com',
    'retries': 0,
    'email_on_retry': False,
    'email_on_failure': False
}


def upload_report(report_params: Dict[str, Any], provider: str, timezone: str, **kwargs) -> None:
    try:
        task = kwargs["ti"]
        report_path = task.xcom_pull(task_ids="CREATE-REPORT", key="REPORT_PATH")
        pattern = report_params["datetime_pattern"]
        conn = report_params["s3_params"]["connection_id"]
        bucket = report_params["s3_params"]["bucket"]
        prefix = report_params["s3_params"]["prefix"]
        key = report_params["s3_params"]["filename"]
        if not prefix.endswith("/"):
            key = prefix + "/" + key
        else:
            key = prefix + key
        upload = load_report_to_s3(conn, bucket, key, report_path, provider, timezone, pattern)
        if upload:
            logger.info(f"Se subio correctamente el reporte a {key}. bucket: {bucket}")
            delete_temp_dir(report_path)
    except Exception as e:
        logger.error(f"Error subiendo reporte a . {e}")


def create_report(tmp_path: str, **kwargs) -> None:
    try:
        task = kwargs["ti"]
        data = task.xcom_pull(task_ids="GET-DATA-REPORT", key="REPORT-DATA")
        status, execution_date = data["PROCESS_STATUS"], data["PROCESS_EXECUTION"]
        _, _ = data.pop("PROCESS_STATUS"), data.pop("PROCESS_EXECUTION")
        dir_name = str(uuid.uuid4())
        if not tmp_path.endswith("/"):
            tmp_path += "/"
        path_dir = tmp_path + dir_name
        os.makedirs(path_dir)
        excel_tmp_path = path_dir + "/tmp_excel.xlsx"
        with pd.ExcelWriter(excel_tmp_path, engine="xlsxwriter") as writer:
            workbook = writer.book
            worksheet = workbook.add_worksheet("report")
            worksheet.set_zoom(90)

            title = "Reporte de último proceso ejecutado"
            title_format = workbook.add_format()
            title_format.set_font_size(20)
            title_format.set_font_color("#333333")

            header = f"Reporte ejecutado el día {execution_date}"
            if status == ProcessStatusEnum.SUCCESS.value:
                status = "EXITOSO"
            elif status == ProcessStatusEnum.FAIL.value:
                status = "FALLIDO"
            elif status == ProcessStatusEnum.RESET.value:
                status = "RESETEADO POR EL USUARIO"
            status = f"Estado de último proceso ejecutado: {status}"
            header_format = workbook.add_format()
            header_format.set_font_size(10)
            header_format.set_font_color("#080606")

            worksheet.merge_range('A1:N1', title, title_format)
            worksheet.merge_range('A2:N2', header, header_format)
            worksheet.merge_range('A3:N3', status, header_format)

            row_format = workbook.add_format()
            row_format.set_font_size(8)
            row_format.set_font_color("#000000")

            base_index = 5
            for index, key in enumerate(data.keys()):
                index = base_index + index
                worksheet.merge_range('A'+str(index)+':B'+str(index), key, row_format)
                if data[key]["TYPE"] == "EXTRACTION":
                    worksheet.merge_range('C'+str(index)+':G'+str(index), f"TABLA DE EXTRACCIÓN: {data[key]['DESCRIPTION']}", row_format)
                elif data[key]["TYPE"] == "TRANSFORMATION":
                    script = data[key]["DESCRIPTION"].split("|")[1]
                    worksheet.merge_range('C'+str(index)+':G'+str(index), f"SCRIPT DE TRANSFORMACIÓN: {script}", row_format)
                elif data[key]["TYPE"] == "GENERATION":
                    worksheet.merge_range('C'+str(index)+':G'+str(index), f"ARCHIVO GENERADO DESDE LA TABLA: {data[key]['DESCRIPTION']}", row_format)
                worksheet.merge_range('H'+str(index)+':I'+str(index), f"ESTADO: {data[key]['STATUS']}", row_format)
                worksheet.merge_range('J'+str(index)+':N'+str(index), data[key]['MESSAGE'], row_format)
            task.xcom_push(key="REPORT_PATH", value=excel_tmp_path)
    except Exception as e:
        logger.error(f"Error creando reporte. {e}")


def get_data_report(**kwargs) -> None:
    try:
        report_data = {}
        task = kwargs['ti']
        control_config = task.xcom_pull(task_ids="CONTROL-EXTRACTOR", key="CONTROL-CONFIG")
        control_key, control = control_config[0], control_config[1]
        if not control:
            logger.info("Archivo de control no encontrado. No se actualizará ningún archivo de control")
        else:
            last_process = control[-1]
            if "reset_by_user" in last_process.keys():
                report_data["PROCESS_EXECUTION"] = ProcessStatusEnum.RESET.value
            else:
                total_tasks = [last_process["tasks"]]
                current_status = last_process["status"]
                control.reverse()
                for process in control:
                    if process["status"] == ProcessStatusEnum.SUCCESS.value:
                        break
                    total_tasks.append(process["tasks"])
                final_key_tasks, final_key_desc, final_key_message = {}, {}, {}
                for tasks in total_tasks:
                    for key in tasks.keys():
                        this_status = tasks[key]["status"]
                        this_desc = tasks[key]["description"]
                        this_message = tasks[key]["message"]
                        if key in final_key_tasks.keys():
                            task_status = final_key_tasks[key]
                            if this_status == ProcessStatusEnum.SUCCESS.value and \
                                    task_status == ProcessStatusEnum.FAIL.value:
                                final_key_tasks.update({key: this_status})
                                final_key_desc.update({key: this_desc})
                                final_key_message.update({key: ''})
                        else:
                            final_key_tasks.update({key: this_status})
                            final_key_desc.update({key: this_desc})
                            final_key_message.update({key: this_message})
                for item in final_key_tasks.keys():
                    if item.lower().startswith("extract"):
                        type_task = "EXTRACTION"
                    elif item.lower().startswith("transform"):
                        type_task = "TRANSFORMATION"
                    else:
                        type_task = "GENERATION"
                    report_data.update({item: {"STATUS": final_key_tasks[item], "TYPE": type_task,
                                        "DESCRIPTION": final_key_desc[item], 'MESSAGE': final_key_message[item]}})
                report_data.update({"PROCESS_STATUS": current_status, "PROCESS_EXECUTION": last_process["date"]})
                task.xcom_push(key="REPORT-DATA", value=report_data)
            logger.info(f"Diccionario de datos para el reporte: {report_data}")
    except Exception as e:
        logger.error(f"Error general creando reporte. {e}")


def set_dag():
    """ DAG that reset the last process Airflow have"""

    import yaml
    from yaml.loader import SafeLoader

    # Cambiar conf_path dependiendo del ambiente, en prod usando k8 y contenedores usar /opt/airflow/dags/app_conf.yml
    # En desarrollo, cualquiera que apunte a su carpeta dags
    conf_path = MAIN_PATH + "dag_conf.yml"
    with open(conf_path) as f:
        data = yaml.load(f, Loader=SafeLoader)
        logger.info(f"CONFIGURACIÓN: {data}")
        conf = data["app"]
    with DAG(DAG_NAME, default_args=DEFAULT_ARGS, description="Proceso que informa del último proceso ejecutado",
             schedule_interval=conf["inform_dag_schedule"], tags=["DAG BCOM - INFORM PROCESS"], catchup=True) as dag:
        control_s3 = conf["control"]["s3_params"]
        timezone = conf["timezone"]
        control_extractor = PythonOperator(
            task_id="CONTROL-EXTRACTOR",
            python_callable=extract_last_control,
            op_kwargs={'conn_id': control_s3["connection_id"], 'bucket': control_s3["bucket"],
                       'prefix': control_s3["prefix"], 'provider': conf["cloud_provider"], 'timezone': timezone},
            trigger_rule="all_success"
        )

        create = PythonOperator(
            task_id="GET-DATA-REPORT",
            python_callable=get_data_report,
            op_kwargs={},
            trigger_rule="all_success"
        )
        tmp_dir = conf["outputs"]["tmp_path"]
        report = PythonOperator(
            task_id="CREATE-REPORT",
            python_callable=create_report,
            op_kwargs={'tmp_path': tmp_dir},
            trigger_rule="all_success"
        )
        report_params = conf["report"]
        upload = PythonOperator(
            task_id="UPLOAD-REPORT",
            python_callable=upload_report,
            op_kwargs={'report_params': report_params, 'provider': conf["cloud_provider"], 'timezone': timezone},
            trigger_rule="all_success"
        )
        control_extractor >> create >> report >> upload
    return dag


globals()["0"] = set_dag()

