from datetime import datetime
from typing import Any, Dict

from airflow import DAG
import json
from airflow.operators.python import PythonOperator

from components.Databases.Database import Database
from components.DatabaseOperation.DatabaseTransformation import delete_table
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from components.S3Route import upload_file
from components.Control import extract_last_control
from components.Utils import delete_temp_dirs

import logging

logger = logging.getLogger()


DAG_NAME = "RESET_PROCESS"

# Change this path if is deployed in prod or dev
MAIN_PATH = "/root/airflow/dags/"

DEFAULT_ARGS = {
    'owner': 'BCOM',
    "start_date": datetime(2023, 7, 28, 22, 9),
    'depends_on_past': False,
    'email': 'caguirre@bytesw.com',
    'retries': 0,
    'email_on_retry': False,
    'email_on_failure': False
}


def update_control(control_params: Dict[str, Any], provider: str, **kwargs) -> None:
    try:
        task = kwargs['ti']
        control_config = task.xcom_pull(task_ids="CONTROL-EXTRACTOR", key="CONTROL-CONFIG")
        control_key, control = control_config[0], control_config[1]
        if not control:
            logger.info("Archivo de control no encontrado. No se actualizará ningún archivo de control")
        else:
            last_process = control[-1]
            last_process.update({'reset_by_user': True})
            control.pop(-1)
            control.append(last_process)
            control = json.dumps(control).encode('utf-8')
            load = upload_file(control, control_params["connection_id"], control_params["bucket"],
                               control_key, provider)
            if load:
                logger.info(f"Subido correctamente el archivo de control en bucket {control_params['bucket']} y"
                            f"key: {control_key}")
        # Borrar las variables que hayan quedado
        delete_all_xcom_tasks()
        delete_task_instances()
    except Exception as e:
        logger.error(f"Error actualizando archivo de control. {e}")


def reset_process(intern_db, output_tmp_dir: str) -> None:
    try:
        # Borrrando tablas
        tablenames = intern_db.get_all_tablenames()
        if len(tablenames) == 0:
            logger.info("No se encontraron tablas para su limpieza")
        else:
            for tablename in tablenames:
                tablename = tablename[0]
                delete = delete_table(tablename, intern_db.engine)
                if delete:
                    logger.info(f"Borrado correctamente la tabla {tablename}")
        # Borrando archivos temporales que hayan quedado sueltos
        delete = delete_temp_dirs(output_tmp_dir)
        if delete:
            logger.info("Se borraron todos los archivos temporales")
    except Exception as e:
        logger.error(f"Error procesando archivo de control. {e}")


def set_dag():
    """ DAG that reset the last process Airflow have"""

    import yaml
    from yaml.loader import SafeLoader

    # Cambiar conf_path dependiendo del ambiente, en prod usando k8 y contenedores usar /opt/airflow/dags/app_conf.yml
    # En desarrollo, cualquiera que apunte a su carpeta dags
    conf_path = MAIN_PATH + "dag_conf.yml"
    with open(conf_path) as f:
        data = yaml.load(f, Loader=SafeLoader)
        logger.info(f"CONFIGURACIÓN: {data}")
        conf = data["app"]
    with DAG(DAG_NAME, default_args=DEFAULT_ARGS, description="Proceso que resetea el último proceso ejecutado",
             schedule_interval=conf["reset_dag_schedule"], tags=["DAG BCOM - RESET PROCESS"], catchup=True) as dag:
        control_s3 = conf["control"]["s3_params"]
        timezone = conf["timezone"]
        control_extractor = PythonOperator(
            task_id="CONTROL-EXTRACTOR",
            python_callable=extract_last_control,
            op_kwargs={'conn_id': control_s3["connection_id"], 'bucket': control_s3["bucket"],
                       'prefix': control_s3["prefix"], 'provider': conf["cloud_provider"], 'timezone': timezone},
            trigger_rule="all_success"
        )

        # Intern Database configuration.
        intern_params = conf["database"]["transformation"]
        intern_db = Database(intern_params["type"], intern_params["host"], int(intern_params["port"]),
                             intern_params["username"], intern_params["password"], intern_params["database"],
                             intern_params["service"], intern_params["schema"])
        intern_db.create_engine()
        tmp_dir = conf["outputs"]["tmp_path"]
        control_process = PythonOperator(
            task_id="RESET-PROCESS",
            python_callable=reset_process,
            op_kwargs={'intern_db': intern_db, 'output_tmp_dir': tmp_dir},
            trigger_rule="all_success"
        )

        update = PythonOperator(
            task_id="UPDATE-CONTROL",
            python_callable=update_control,
            op_kwargs={'control_params': control_s3, 'provider': conf["cloud_provider"]},
            trigger_rule="all_success"
        )

        control_extractor >> control_process >> update

    return dag


globals()["0"] = set_dag()

